# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module to act as a test north endpoint, logging its inputs """

import os
import copy
import asyncio

import csv

import logging
import datetime

from foglamp.common import logger

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"



_DEFAULT_CONFIG = {
    'plugin': {        # mandatory
        'description': 'north plugin which writes its output to a series of csv files',
        'type': 'string',
        'default': 'csv_writer',
        'readonly': 'true'
    },
    'enabled': {	    # mandatory
        'description': 'Enable / disable this plugin',
        'type': 'boolean',
        'default': "False",
        'displayName': 'Enable this plugin',
        'order': '1'
    },
    "source": {		# mandatory north
        "description": "The source of data to send",
        "type": "enumeration",
        "default": "readings",
        "displayName": "Data Source",
        "options": ["readings", "statistics"],
        'order': '2',
    },
    "applyFilter": {
        "description": "Whether to apply filter before processing the data",
        "type": "boolean",
        "default": "false",
        "readonly": "True",
        "order": '3',
        "displayName": "Apply Filter"
    },
    "filterRule": {
        "description": "JQ formatted filter to apply (applicable if applyFilter is True)",
        "type": "string",
        "readonly": "True",
        "default": ".[]",
        "order": '4',
        "displayName": "Filter Rule",
        "validity": "applyFilter == \"true\""
    },

    # csv writer-specfic values
    'maxRecords': {
        'description': 'Max records written per file',
        'type': 'integer',
        'default': '100000',
        'displayName': 'Max records',
        'order': '5'
    },
    'maxFiles': {
        'description': 'Max files accumulated before rotation',
        'type': 'integer',
        'default': '100',
        'displayName': 'Max rotation',
        'order': '6'
    },
    'destDir': {
        'description': 'Destination directory',
        'type': 'enumeration',
        'default': 'FOGLAMP_DATA/csv',
        'options': "['FOGLAMP_DATA/csv', '/media/csv']",
        'displayName': 'Destination directory',
        'order': '7'
    },
}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("CSVW - plugin_info called")
    return {
        'name': 'CSV Writer',
        'version': '1.9.0',
        'type': 'north',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


# local logger
_LOGGER = logger.setup(__name__, level=logging.INFO)

# static handle to the (transient) state maintained by this plugin
_csv_handle = None

class CSVHandle(object):
    """
    Stores state passed to/from FogLAMP callbacks to manage periodic forwarding of data
    """

    def __init__(self, source, dest_dir, max_records, max_files):

        self.source = source   # readings/audit/...
        
        if dest_dir.startswith('FOGLAMP_DATA/'):
            FOGLAMP_DATA = os.getenv("FOGLAMP_DATA", default="/usr/local/foglamp/data")
            self.dest_dir = "{}/{}".format(FOGLAMP_DATA, dest_dir[len("FOGLAMP_DATA/"):])
        else:
            self.dest_dir = _dest_dir

        if not os.path.exists(self.dest_dir):
            os.makedirs(self.dest_dir)

        self.max_records = max_records
        self.max_files = max_files

        
        # ### state ###
        self.csv = None
        self.n_records = self.max_records  # force opening a new file
        self.n_cur_file = 0
        self.ext = 0

        self._clean_dir()

        
    def _clean_dir(self):
        """ clean_dir - remove all data files in the csv target area """
        for f in filter(lambda f: f.endswith('.csv'), os.listdir(self.dest_dir)):
            try:
                os.remove("{}/{}".format(self.dest_dir, f))
            except:
                _LOGGER.info("failed removing {}".format(f))
                             
    def _new_csv(self, fields):
        """ rotate_dir - delete the oldest file and open a new one """
        _files = sorted([f for f in os.listdir(self.dest_dir) if f.endswith('.csv')])
        if len(_files) >= self.max_files:
            # if we have the max files possible, rotate
            _rm_file = "{}/{}".format(self.dest_dir, _files[0])
            _LOGGER.info("NEW - ROTATE: removing {}".format(_rm_file))
            os.remove(_rm_file)

        _hms = datetime.datetime.now().strftime("%H-%M-%S")
        _cur_file = "{}/{}-0000.csv".format(self.dest_dir, _hms)
        _LOGGER.info("NEW - new file {}".format(_cur_file))
        if os.path.exists(_cur_file):
            # rotating in under a second -- uniquify it
                         
            self.ext += 1
            _cur_file = "{}/{}-{:04}.csv".format(self.dest_dir, _hms, self.ext)
            _LOGGER.info("NEW - overlap iterate file {}".format(_cur_file))
        else:
            self.ext = 0

        self.csv = csv.DictWriter(open(_cur_file, 'w'), fieldnames=fields)
        self.csv.writeheader()
        self.n_records = 0

        
    def save(self, data):

        _len = 0
        for r in data:
            # _LOGGER.info('CSVW: ', data)
            if self.n_records == self.max_records:
                # when we've filled the file, start a new one, maybe rotating out an old one
                self._new_csv(r.keys())

            _len += self.csv.writerow(r)
            self.n_records += 1

        return _len
        


def plugin_init(config):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    _LOGGER.info("CSVW - plugin_init called")
    try:
        global _csv_handle

        _config = config # copy.deepcopy(config)

        _source = _config['source']['value']
        _dest_dir = _config['destDir']['value']
        _max_records = int(_config['maxRecords']['value'])
        _max_files =int(_config['maxFiles']['value'])

        # store configuration options and working state
        _csv_handle = CSVHandle(_source, _dest_dir, _max_records, _max_files)

    except:
        _LOGGER.info("could not create configuration")
        raise

    return _config

async def plugin_send(handle, data, input_ref):
    """ plugin_send -- log data we receive """

    try:
        _data_sent = False
        _len = 0
        _offset = 0

        # print("CSVW send: ", data)
        if type(data) == dict:
            data = [data]
        
        if len(data) > 0:
            # readings / statistics / assets  - depending, there may be a trailing 's' or not
            _source = _csv_handle.source

            if _source in data[0]:
                _idx = _source
            elif _source+'s' in data[0]:
                _idx = _source+'s'
            else:
                _LOGGER.error("missing data field {}".format(_source))
                raise ValueError
                
            _len = _csv_handle.save([d[_idx] for d in data])
            _data_sent = True
            _offset = 0

    except:
        _LOGGER.error("csv writer exception")
        raise

    return _data_sent, _offset, _len

    
def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
        INAPPLICABLE for north plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
    """
    _LOGGER.info("DBGL - reconfigure")
    pass


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
        INAPPLICABLE for north plugin
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
    """
    _LOGGER.info("DBGL - SHUTDOWN")
    pass
